# -*- coding: utf-8 -*-
# 2025/9/1
# create by: snower

import argparse
import hashlib
import logging
import os
import struct
import sys
import time
import traceback
import weakref
import zlib
from collections import deque

import sevent
from sevent.helpers.tcp_forward import tcp_forward
from sevent.helpers.simple_proxy import parse_allow_deny_host, parse_allow_deny_host_filename, check_timeout, tcp_proxy, \
    warp_write
from sevent.helpers.utils import config_signal, create_server, create_socket, format_data_len
from ..buffer import Buffer, BaseBuffer
from ..errors import SocketClosed, ConnectError
from ..event import EventEmitter, null_emit_callback
from ..loop import instance
from ..tcp import Socket, STATE_STREAMING, STATE_CLOSED, STATE_CLOSING, STATE_INITIALIZED
from ..utils import get_logger

FRAME_TYPE_AUTH = 0x01
FRAME_TYPE_AUTHED = 0x02
FRAME_TYPE_PING = 0x03
FRAME_TYPE_PONG = 0x04
FRAME_TYPE_OPEN = 0x11
FRAME_TYPE_CLOSING = 0x12
FRAME_TYPE_CLOSED = 0x13
FRAME_TYPE_RESET = 0x14
FRAME_TYPE_DATA = 0x15
FRAME_TYPE_DRAIN = 0x16
FRAME_TYPE_REGAIN = 0x017

FRAME_MAX_SIZE = 2560


class TunnelStream(Socket):
    def __init__(self, loop=None, stream_id=None, tunnel=None, max_buffer_size=None, compressor=None, decompressor=None):
        EventEmitter.__init__(self)

        self._loop = loop or instance()
        self._stream_id = stream_id
        self._tunnel = tunnel
        self._compressor = compressor
        self._decompressor = decompressor
        self._max_buffer_size = max_buffer_size or Socket.MAX_BUFFER_SIZE
        self._rbuffers = Buffer(max_buffer_size=self._max_buffer_size)
        self._wbuffers = Buffer(max_buffer_size=self._max_buffer_size)
        self._state = STATE_STREAMING if tunnel else STATE_INITIALIZED
        self._has_drain_event = False
        self._writing = False
        self.ignore_write_closed_error = False
        self._frame_closing = False
        self._recv_drain_waiting_regain = False
        if self._state == STATE_STREAMING:
            self._rbuffers.on("drain", lambda _: self.drain())
            self._rbuffers.on("regain", lambda _: self.regain())

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def address(self):
        return "tunnel#%d" % id(self._tunnel), self._stream_id

    @property
    def socket(self):
        return self

    def enable_fast_open(self):
        pass

    @property
    def is_enable_fast_open(self):
        return False

    def enable_nodelay(self):
        pass

    @property
    def is_enable_nodelay(self):
        return True

    def end(self):
        if self._state != STATE_STREAMING:
            return
        if self._wbuffers:
            self._state = STATE_CLOSING
        else:
            self._loop.add_async(self.close)

    def close(self):
        if self._state == STATE_CLOSED:
            return
        self._state = STATE_CLOSED
        if not self._frame_closing:
            self._wbuffers.read()
            try:
                self._tunnel.write_frame(self._stream_id, FRAME_TYPE_CLOSING, 0, None)
            except Exception as e:
                get_logger().exception("write closing frame error:%s", e)
                self.do_close()
        else:
            self._wbuffers.read()
            try:
                self._tunnel.write_frame(self._stream_id, FRAME_TYPE_CLOSED, 0, None)
            except Exception as e:
                get_logger().exception("write closed frame error:%s", e)
            self.do_close()

    def do_end(self):
        self._frame_closing = True
        if self._state != STATE_STREAMING:
            return
        if self._wbuffers:
            self._state = STATE_CLOSING
            if self._recv_drain_waiting_regain:
                self._recv_drain_waiting_regain = False
                if not self._writing and self._wbuffers:
                    self.do_write_drain()
        else:
            self._tunnel.write_frame(self._stream_id, FRAME_TYPE_CLOSED, 0, None)
            self.do_close()

    def do_close(self):
        if self._rbuffers is None or self._wbuffers is None:
            return
        self._state = STATE_CLOSED

        def on_close():
            if self._rbuffers is None or self._wbuffers is None:
                return
            self._tunnel.close_stream(self)
            try:
                self.emit_close(self)
            except Exception as e:
                get_logger().exception("emit close error:%s", e)
            self.remove_all_listeners()
            self._rbuffers.close()
            self._wbuffers.close()
            self._rbuffers = None
            self._wbuffers = None
        self._loop.add_async(on_close)

    def _error(self, error):
        self._loop.add_async(self.emit_error, self, error)
        self._loop.add_async(self.close)
        if self.emit_error == null_emit_callback:
            get_logger().error("TcpTunnelStream %s socket %s error: %s", self, self.socket, error)

    def connect(self, address, timeout=5):
        if self._state != STATE_INITIALIZED:
            return
        tunnel = TcpTunnel.get_tunnel(address[0])
        if not tunnel:
            raise ConnectError(None, None)
        self._tunnel.connect_stream(self)
        self._state = STATE_STREAMING
        self._rbuffers.on("drain", lambda _: self.drain())
        self._rbuffers.on("regain", lambda _: self.regain())
        self._loop.add_async(self.emit_connect, self)

    def drain(self):
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            self._tunnel.write_frame(self._stream_id, FRAME_TYPE_DRAIN, 0, None)

    def regain(self):
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            self._tunnel.write_frame(self._stream_id, FRAME_TYPE_REGAIN, 0, None)

    def do_on_drain(self):
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            self._recv_drain_waiting_regain = True
            self._wbuffers.do_drain()

    def do_on_regain(self):
        self._recv_drain_waiting_regain = False
        if self._state in (STATE_STREAMING, STATE_CLOSING):
            if not self._writing and self._wbuffers:
                self._writing = True
                self.do_write_drain()
            else:
                self._wbuffers.do_regain()

    def do_on_frame(self, frame_type, frame_flag, data):
        if not data or self._rbuffers is None:
            return
        if self._decompressor is not None:
            data = self._decompressor.decompress(data)
            if not data:
                return
        BaseBuffer.write(self._rbuffers, data)
        if self._rbuffers._len > self._rbuffers._drain_size and not self._rbuffers._full:
            self._rbuffers.do_drain()
        self._loop.add_async(self.emit_data, self, self._rbuffers)

    def do_write_drain(self):
        if self._state not in (STATE_STREAMING, STATE_CLOSING):
            return
        if not self._recv_drain_waiting_regain and self._wbuffers:
            try:
                data = BaseBuffer.read(self._wbuffers, FRAME_MAX_SIZE) if len(self._wbuffers) > FRAME_MAX_SIZE else BaseBuffer.read(self._wbuffers)
                self._tunnel.write_frame(self._stream_id, FRAME_TYPE_DATA, 0, data)
                if self._wbuffers._full and self._wbuffers._len < self._wbuffers._regain_size:
                    self._wbuffers.do_regain()
                if not self._wbuffers and self._has_drain_event:
                    self._loop.add_async(self.emit_drain, self)
            except Exception as e:
                self._writing = False
                self._error(e)
        else:
            self._writing = False
            if self._state == STATE_CLOSING:
                self.close()

    def write(self, data):
        if self._state != STATE_STREAMING:
            if self.ignore_write_closed_error:
                return False
            raise SocketClosed()
        if data.__class__ is Buffer:
            if self._compressor is not None:
                if not data:
                    return True if not self._wbuffers else False
                BaseBuffer.write(self._wbuffers, self._compressor.compress(data.read()))
                BaseBuffer.write(self._wbuffers, self._compressor.flush(zlib.Z_SYNC_FLUSH))
            else:
                BaseBuffer.extend(self._wbuffers, data)
            if data._full and data._len < data._regain_size:
                data.do_regain()
        else:
            if self._compressor is not None:
                BaseBuffer.write(self._wbuffers, self._compressor.compress(data))
                BaseBuffer.write(self._wbuffers, self._compressor.flush(zlib.Z_SYNC_FLUSH))
            else:
                BaseBuffer.write(self._wbuffers, data)
        if self._wbuffers._len > self._wbuffers._drain_size and not self._wbuffers._full:
            self._wbuffers.do_drain()
        if not self._writing and not self._recv_drain_waiting_regain:
            if not self._wbuffers:
                return True
            data = BaseBuffer.read(self._wbuffers, FRAME_MAX_SIZE) if len(self._wbuffers) > FRAME_MAX_SIZE else BaseBuffer.read(self._wbuffers)
            self._tunnel.write_frame(self._stream_id, FRAME_TYPE_DATA, 0, data)
            self._writing = True
            if self._wbuffers._full and self._wbuffers._len < self._wbuffers._regain_size:
                self._wbuffers.do_regain()
            if not self._wbuffers:
                if self._has_drain_event:
                    self._loop.add_async(self.emit_drain, self)
                return True
        return False


class TcpTunnel(EventEmitter):
    _tunnels = weakref.WeakValueDictionary()

    @classmethod
    def get_tunnel(cls, tunnel_id):
        if isinstance(tunnel_id, str) and tunnel_id.startswith("tunnel#"):
            tunnel_id = int(tunnel_id[7:])
        return cls._tunnels.get(tunnel_id)

    def __init__(self, loop=None, is_server=False):
        super(TcpTunnel, self).__init__()

        self._loop = loop or instance()
        self._is_server = is_server
        self._socket = None
        self._compress_factory = None
        self._streams = {}
        self._current_id_index = 1 if is_server else 2
        self._send_queue = deque()
        self._send_waiting_drain = False
        self._send_timestamp = 0
        self._recv_length = 2
        self._recv_waiting_length = True
        self._recv_timestamp = 0
        self._ping_timestamp = time.time()
        self._pong_timestamp = time.time()
        self._tunnels[id(self)] = self

    @property
    def streams(self):
        return self._streams

    @property
    def streams_count(self):
        return len(self._streams)

    @property
    def address(self):
        if self._socket is None:
            return "", 0
        return self._socket.address

    def update_socket(self, socket, compress_factory):
        self._socket = socket
        self._compress_factory = compress_factory
        self._send_queue.clear()
        self._send_waiting_drain = False
        self._recv_length = 2
        self._recv_waiting_length = True
        socket.on_data(self.on_data)
        socket.on_close(self.on_close)
        socket.on_drain(self.on_drain)
        self._ping_timestamp = time.time()
        self._pong_timestamp = time.time()
        def do_ping_timeout():
            if self._socket is not socket:
                return
            now = time.time()
            if now - self._recv_timestamp <= 45:
                self._ping_timestamp = self._send_timestamp
                self._pong_timestamp = self._recv_timestamp
                self._loop.add_timeout(20 if self._is_server else 15, do_ping_timeout)
                return
            if now - self._pong_timestamp > 120:
                socket.close()
                return
            try:
                if now - self._ping_timestamp >= 45:
                    self.write_frame(0, FRAME_TYPE_PING, 0, None)
                    self._ping_timestamp = now
            finally:
                self._loop.add_timeout(20 if self._is_server else 15, do_ping_timeout)
        self._loop.add_timeout(20 if self._is_server else 15, do_ping_timeout)

    def open_stream(self):
        if self._socket is None:
            raise ConnectError(None, None)
        while True:
            stream_id = self._current_id_index
            self._current_id_index += 2
            if self._current_id_index > 0xffff:
                self._current_id_index = 1 if self._is_server else 2
            if stream_id not in self._streams:
                break
        stream = TunnelStream(self._loop, stream_id, self, compressor=self._compress_factory.create_compressor(),
                              decompressor=self._compress_factory.create_decompressor())
        self._streams[stream_id] = stream
        self.write_frame(stream_id, FRAME_TYPE_OPEN, 0, None)
        return stream

    def connect_stream(self, stream):
        if self._socket is None:
            raise ConnectError(None, None)
        while True:
            stream_id = self._current_id_index
            self._current_id_index += 2
            if self._current_id_index > 0xffff:
                self._current_id_index = 1 if self._is_server else 2
            if stream_id not in self._streams:
                break
        stream._stream_id = stream_id
        stream._tunnel = self
        self._streams[stream_id] = stream
        self.write_frame(stream_id, FRAME_TYPE_OPEN, 0, None)
        return stream

    def close_stream(self, stream):
        if stream.stream_id not in self._streams:
            return
        self._streams.pop(stream.stream_id, None)

    def write_frame(self, stream_id, frame_type, frame_flag, data):
        if self._socket is None:
            raise SocketClosed()
        if self._send_waiting_drain:
            self._send_queue.append((stream_id, frame_type, frame_flag, data))
        else:
            if data is not None:
                self._socket.write(struct.pack(">HHBB", len(data) + 4, stream_id, frame_type, frame_flag) + data)
            else:
                self._socket.write(struct.pack(">HHBB", 4, stream_id, frame_type, frame_flag))
            self._send_waiting_drain = True
            self._send_timestamp = time.time()
            if frame_type == FRAME_TYPE_DATA:
                stream = self._streams.get(stream_id)
                if stream is not None:
                    self._loop.add_async(stream.do_write_drain)

    def on_data(self, socket, buffer):
        while len(buffer) >= self._recv_length:
            if self._recv_waiting_length:
                self._recv_waiting_length = False
                self._recv_length, = struct.unpack(">H", buffer.read(2))
            else:
                stream_id, frame_type, frame_flag = struct.unpack(">HBB", buffer.read(4))
                data = buffer.read(self._recv_length - 4) if self._recv_length > 4 else None
                self._recv_length = 2
                self._recv_waiting_length = True
                if stream_id == 0:
                    self.on_system_frame(frame_type, frame_flag, data)
                else:
                    self.on_frame(stream_id, frame_type, frame_flag, data)
                self._recv_timestamp = time.time()

    def on_frame(self, stream_id, frame_type, frame_flag, data):
        if stream_id not in self._streams:
            if self._socket is None:
                return
            if frame_type == FRAME_TYPE_OPEN:
                stream = TunnelStream(self._loop, stream_id, self, compressor=self._compress_factory.create_compressor(),
                                      decompressor=self._compress_factory.create_decompressor())
                self._streams[stream_id] = stream
                self._loop.add_async(self.emit_stream, self, stream)
            elif frame_type != FRAME_TYPE_RESET:
                self.write_frame(stream_id, FRAME_TYPE_RESET, 0, None)
            return
        stream = self._streams[stream_id]
        if frame_type == FRAME_TYPE_DATA:
            stream.do_on_frame(frame_type, frame_flag, data)
            return
        if frame_type == FRAME_TYPE_DRAIN:
            stream.do_on_drain()
            return
        if frame_type == FRAME_TYPE_REGAIN:
            stream.do_on_regain()
            return
        if frame_type == FRAME_TYPE_RESET or frame_type == FRAME_TYPE_CLOSED:
            stream.do_close()
            self.close_stream(stream)
            return
        if frame_type == FRAME_TYPE_CLOSING:
            stream.do_end()
            return
        if frame_type == FRAME_TYPE_OPEN:
            self.write_frame(stream_id, FRAME_TYPE_RESET, 0, None)
            stream.do_close()
            self.close_stream(stream)

    def on_system_frame(self, frame_type, frame_flag, data):
        if frame_type == FRAME_TYPE_PING:
            self.write_frame(0, FRAME_TYPE_PONG, 0, None)
            self._pong_timestamp = time.time()
            return

    def on_drain(self, socket):
        if not self._send_queue:
            self._send_waiting_drain = False
            return
        try:
            stream_id, frame_type, frame_flag, data = self._send_queue.popleft()
            if data is not None:
                self._socket.write(struct.pack(">HHBB", len(data) + 4, stream_id, frame_type, frame_flag) + data)
            else:
                self._socket.write(struct.pack(">HHBB", 4, stream_id, frame_type, frame_flag))
            self._send_timestamp = time.time()
            if frame_type == FRAME_TYPE_DATA:
                stream = self._streams.get(stream_id)
                if stream is not None:
                    self._loop.add_async(stream.do_write_drain)
        except Exception as e:
            self._send_waiting_drain = False
            if self._socket is not None:
                self._socket._error(e)

    def on_close(self, socket):
        self._socket = None
        streams = list(self._streams.values())
        self._streams.clear()
        self._send_queue.clear()
        self._send_waiting_drain = False
        self._recv_length = 2
        self._recv_waiting_length = True
        for stream in streams:
            stream.do_close()

    def close(self):
        if self._socket:
            self._socket.close()
        if id(self) in self._tunnels:
            self._tunnels.pop(id(self), None)


class CompressFactory(object):
    @classmethod
    def build(cls, compress_method):
        if compress_method is not None and compress_method == 1:
            return ZlibDeflatedCompressFactory()
        return CompressFactory()

    def create_compressor(self):
        return None

    def create_decompressor(self):
        return None


class ZlibDeflatedCompressFactory(CompressFactory):
    def create_compressor(self):
        return zlib.compressobj(
            level=int(os.environ.get("SEVENT_COMPRESS_LEVEL", 6)),
            method=zlib.DEFLATED,
            wbits=-zlib.MAX_WBITS
        )

    def create_decompressor(self):
        return zlib.decompressobj(
            wbits=-zlib.MAX_WBITS
        )


def gen_sign_key(key):
    t = struct.pack("I", int(time.time()))
    oncestr = os.urandom(16)
    return t + oncestr + hashlib.md5(t + oncestr + sevent.utils.ensure_bytes(key)).digest()

def check_sign_key(key, sign_key):
    return sign_key[20:] == hashlib.md5(sign_key[:20] + sevent.utils.ensure_bytes(key)).digest()

async def handler_server_conn(conns, tunnel, conn, key, is_compress_mode, error_forward_address):
    def on_timeout():
        if conn is None or tunnel._socket is conn:
            return
        conn.close()
    timeout_handler = sevent.current().add_timeout(15, on_timeout)

    start_time = time.time()
    try:
        logging.info("remote conn connecting %s:%d", conn.address[0], conn.address[1])
        origin_data = (await conn.recv(2)).read()
        data_length, = struct.unpack(">H", origin_data[:2])
        if data_length <= 6 or data_length > 64 or data_length > len(origin_data):
            if error_forward_address:
                logging.info("remote conn length fail %s:%d %d %d", conn.address[0], conn.address[1], data_length, len(origin_data))
                status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
                conn.buffer[0].write(origin_data + conn.buffer[0].read())
                await tcp_forward(conns, conn, error_forward_address, None, status)
                return
            await conn.send(struct.pack(">HHBBBB", 6, 0, FRAME_TYPE_AUTHED, 0, 1, 0))
            await conn.closeof()
            sevent.current().cancel_timeout(timeout_handler)
            logging.info("remote conn length fail %s:%d %d", conn.address[0], conn.address[1], data_length)
            return
        data = origin_data[2:2 + data_length]
        _, frame_type, frame_flag, sign_key_length = struct.unpack(">HBBH", data[:6])
        sign_key = data[6:6 + sign_key_length]
        compress_method_count = data[6 + sign_key_length] if len(data) > 6 + sign_key_length else -1
        compress_methods = data[7 + sign_key_length:7 + sign_key_length + compress_method_count] if compress_method_count > 0 else None
        if not check_sign_key(key, sign_key):
            if error_forward_address:
                logging.info("remote conn auth fail %s:%d %s", conn.address[0], conn.address[1], sign_key)
                status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
                conn.buffer[0].write(origin_data + conn.buffer[0].read())
                await tcp_forward(conns, conn, error_forward_address, None, status)
                return
            await conn.send(struct.pack(">HHBBBB", 6, 0, FRAME_TYPE_AUTHED, 0, 1, 0))
            await conn.closeof()
            sevent.current().cancel_timeout(timeout_handler)
            logging.info("remote conn auth fail %s:%d %s", conn.address[0], conn.address[1], sign_key)
            return

        if tunnel._socket is not None:
            await conn.send(struct.pack(">HHBBBB", 6, 0, FRAME_TYPE_AUTHED, 0, 2, 0))
            await conn.closeof()
            sevent.current().cancel_timeout(timeout_handler)
            logging.info("remote conn auth fail %s:%d %s", conn.address[0], conn.address[1], sign_key)
            return

        await conn.send(struct.pack(">HHBBBB", 6, 0, FRAME_TYPE_AUTHED, 0, 0, compress_methods[0] if is_compress_mode and compress_methods else 0))
        sevent.current().cancel_timeout(timeout_handler)
        tunnel.update_socket(conn, CompressFactory.build(compress_methods[0] if is_compress_mode and compress_methods else 0))
        logging.info("remote conn succeed %s:%d", conn.address[0], conn.address[1])
        await conn.join()
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("remote conn error %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1],
                     e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        if conn is not None:
            conn.close()
    logging.info("remote conn closed %s:%d %.2fms", conn.address[0], conn.address[1], (time.time() - start_time) * 1000)

async def run_tunnel_server(conns, tunnel, listen_host, listen_port, key, is_compress_mode, error_forward_address):
    server = create_server((listen_host, listen_port))
    while True:
        conn = await server.accept()
        sevent.current().call_async(handler_server_conn, conns, tunnel, conn, key, is_compress_mode, error_forward_address)

async def run_tunnel_client(tunnel, connect_host, connect_port, key, is_compress_mode):
    while True:
        start_time = time.time()
        is_connected, conn = False, None
        try:
            logging.info("local conn connecting -> %s:%d", connect_host, connect_port)
            conn = create_socket((connect_host, connect_port))
            await conn.connectof((connect_host, connect_port))

            def on_timeout():
                if conn is None or tunnel._socket is conn:
                    return
                conn.close()
            timeout_handler = sevent.current().add_timeout(15, on_timeout)

            sign_key = gen_sign_key(key)
            compressed = b'\x01\x01' if is_compress_mode else b'\x00'
            await conn.send(struct.pack("!HHBBH", 6 + len(sign_key) + len(compressed), 0, FRAME_TYPE_AUTH, 0, len(sign_key)) + sign_key + compressed)
            data_length, = struct.unpack(">H", (await conn.recv(2)).read(2))
            data = (await conn.recv(data_length)).read(data_length)
            _, frame_type, frame_flag, connect_result = struct.unpack(">HBBB", data[:5])
            compress_method = data[5] if len(data) >= 6 else 0
            if connect_result != 0:
                await conn.closeof()
                sevent.current().cancel_timeout(timeout_handler)
                logging.info("local conn fail -> %s:%d %d", connect_host, connect_port, connect_result)
            else:
                sevent.current().cancel_timeout(timeout_handler)
                tunnel.update_socket(conn, CompressFactory.build(compress_method))
                is_connected = True
                logging.info("local conn succeed -> %s:%d", connect_host, connect_port)
                await conn.join()
        except sevent.errors.SocketClosed:
            logging.info("local conn closed -> %s:%d %.2fms", connect_host, connect_port, (time.time() - start_time) * 1000)
        except Exception as e:
            logging.info("local conn error -> %s:%d %s %.2fms\r%s", connect_host, connect_port,
                         e, (time.time() - start_time) * 1000, traceback.format_exc())
        else:
            logging.info("local conn closed -> %s:%d %.2fms", connect_host, connect_port, (time.time() - start_time) * 1000)
        finally:
            if conn is not None:
                conn.close()
        await sevent.sleep(3 if not is_connected else 0.1)

async def handle_proxy_local(conns, tunnel, conn, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None
    try:
        pconn = tunnel.open_stream()
        logging.info("tunnel#%d streams %d open stream %d", id(tunnel), tunnel.streams_count, pconn.stream_id)
        try:
            pconn.write = warp_write(pconn, status, "send_len")
            logging.info("connected %s:%s -> %s:%s -> %s:%s", conn.address[0], conn.address[1], tunnel.address[0], tunnel.address[1],
                         pconn.address[0], pconn.address[1])
            await conn.linkof(pconn)
        finally:
            logging.info("tunnel#%d streams %d close stream %d", id(tunnel), tunnel.streams_count, pconn.stream_id)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("error %s:%s -> %s:%s -> %s:%s %s %.2fms\r%s", conn.address[0], conn.address[1],
                     tunnel.address[0], tunnel.address[1], pconn.address[0] if pconn else "", pconn.address[1] if pconn else 0,
                     e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("closed %s:%s -> %s:%s -> %s:%s %s %s %.2fms", conn.address[0], conn.address[1],
                 tunnel.address[0], tunnel.address[1], pconn.address[0], pconn.address[1], format_data_len(status["send_len"]),
                 format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

async def run_proxy_local_server(conns, tunnel, bind_host, bind_port):
    server = create_server((bind_host, bind_port))
    while True:
        conn = await server.accept()
        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
        sevent.current().call_async(handle_proxy_local, conns, tunnel, conn, status)
        conns[id(conn)] = (conn, status)

async def handle_proxy_remote_proxy(conns, tunnel, stream, status, allow_host_names, deny_host_names):
    logging.info("tunnel#%d streams %d open stream %d", id(tunnel), tunnel.streams_count, stream.stream_id)
    try:
        await tcp_proxy(conns, tunnel, stream, status, allow_host_names, deny_host_names)
    finally:
        logging.info("tunnel#%d streams %d close stream %d", id(tunnel), tunnel.streams_count, stream.stream_id)

def handle_proxy_remote_stream(conns, tunnel, stream, allow_host_names, deny_host_names):
    status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
    sevent.current().call_async(handle_proxy_remote_proxy, conns, tunnel, stream, status, allow_host_names, deny_host_names)
    conns[id(stream)] = (stream, status)

def main(argv):
    parser = argparse.ArgumentParser(description='Simple bidirectional http and socks5 proxy based on TCP tunnel')
    parser.add_argument('-c', dest='is_client_mode', nargs='?', const=True, default=False, type=bool, help='is client mode (defualt: False)')
    parser.add_argument('-k', dest='key', default='', type=str, help='auth key (defualt: "")')
    parser.add_argument('-b', dest='bind', default="0.0.0.0", help='proxy bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='port', default=8088, type=int, help='proxy bind port (default: 8088)')
    parser.add_argument('-r', dest='listen_host', default="0.0.0.0", help='server mode reverse server listen host (default: 0.0.0.0)')
    parser.add_argument('-l', dest='listen_port', default=8078, type=int, help='server mode reverse server listen port (default: 8078)')
    parser.add_argument('-H', dest='connect_host', default="127.0.0.1", help='client mode reverse client connect server host (default: 127.0.0.1)')
    parser.add_argument('-P', dest='connect_port', default=8078, type=int, help='client mode reverse client connect server port (default: 8078)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-a', dest='allow_hosts', default=[], action="append", type=str,help='allow forward host name')
    parser.add_argument('-d', dest='deny_hosts', default=[], action="append", type=str, help='deny forward host name')
    parser.add_argument('-A', dest='allow_filename', default="", type=str, help='allow forward host name config filename')
    parser.add_argument('-D', dest='deny_filename', default="", type=str, help='deny forward host name config filename')
    parser.add_argument('-C', dest='is_compress_mode', nargs='?', const=True, default=False, type=bool, help='is client mode (defualt: False)')
    parser.add_argument('-e', dest='error_forward_host', default="", help='error key forward host, accept format [proxy_host:proxy_port]')
    args = parser.parse_args(args=argv)
    config_signal()

    allow_host_names, deny_host_names = [], []
    if args.allow_hosts:
        for allow_host in args.allow_hosts:
            allow_host_names.extend([parse_allow_deny_host(h) for h in allow_host.split(",") if h.strip()])
    if args.deny_hosts:
        for deny_host in args.deny_hosts:
            deny_host_names.extend([parse_allow_deny_host(h) for h in deny_host.split(",") if h.strip()])
    if args.allow_filename:
        allow_host_names.extend(parse_allow_deny_host_filename(args.allow_filename))
    if args.deny_filename:
        deny_host_names.extend(parse_allow_deny_host_filename(args.deny_filename))
    if args.error_forward_host:
        error_forward_info = args.error_forward_host.split(":")
        if len(error_forward_info) == 1:
            if not error_forward_info[0].isdigit():
                error_forward_address = (error_forward_info[0], 80)
            else:
                error_forward_address = ("127.0.0.1", int(error_forward_info[0]))
        else:
            error_forward_address = (error_forward_info[0], int(error_forward_info[1]))
    else:
        error_forward_address = None

    conns = {}
    tunnel = TcpTunnel(is_server=not args.is_client_mode)
    if args.is_client_mode:
        sevent.current().call_async(run_tunnel_client, tunnel, args.connect_host, args.connect_port, args.key, args.is_compress_mode)
        logging.info("listen server at %s:%d -> %s:%s", args.bind, args.port, args.connect_host, args.connect_port)
    else:
        sevent.current().call_async(run_tunnel_server, conns, tunnel, args.listen_host, args.listen_port, args.key, args.is_compress_mode, error_forward_address)
        logging.info("listen server at %s:%d -> %s:%s", args.bind, args.port, args.listen_host, args.listen_port)
    tunnel.on("stream", lambda _, stream: handle_proxy_remote_stream(conns, tunnel, stream,
                                                                     allow_host_names or None, deny_host_names or None))

    sevent.current().call_async(run_proxy_local_server, conns, tunnel, args.bind, args.port)
    sevent.current().call_async(check_timeout, conns, args.timeout)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-

from collections import defaultdict
from .utils import is_py3, get_logger


def null_emit_callback(*args, **kwargs):
    return None


class EventEmitter(object):
    def __init__(self):
        self._events = defaultdict(set)
        self._events_once = defaultdict(set)

    def on(self, event_name, callback):
        event_callbacks = self._events[event_name]
        event_callbacks.add(callback)

        if not self._events_once[event_name] and len(event_callbacks) == 1:
            setattr(self, "emit_" + event_name, callback)
        else:
            setattr(self, "emit_" + event_name, self.emit_callback(event_name))

    def off(self, event_name, callback):
        event_callbacks = self._events[event_name]
        try:
            event_callbacks.remove(callback)
        except KeyError:
            pass

        if not event_callbacks:
            once_event_callbacks = self._events_once[event_name]
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, null_emit_callback)
            elif len(once_event_callbacks) == 1:
                callback = list(once_event_callbacks)[0]

                def emit_callback(*args, **kwargs):
                    setattr(self, "emit_" + event_name, null_emit_callback)
                    once_event_callbacks.clear()
                    return callback(*args, **kwargs)

                setattr(self, "emit_" + event_name, emit_callback)
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        elif len(event_callbacks) == 1:
            once_event_callbacks = self._events_once[event_name]
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, list(event_callbacks)[0])
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        else:
            setattr(self, "emit_" + event_name, self.emit_callback(event_name))

    def once(self, event_name, callback):
        once_event_callbacks = self._events_once[event_name]
        once_event_callbacks.add(callback)

        if not self._events[event_name] and len(once_event_callbacks) == 1:
            def emit_callback(*args, **kwargs):
                setattr(self, "emit_" + event_name, null_emit_callback)
                once_event_callbacks.clear()
                return callback(*args, **kwargs)

            setattr(self, "emit_" + event_name, emit_callback)
        else:
            setattr(self, "emit_" + event_name, self.emit_callback(event_name))

    def noce(self, event_name, callback):
        once_event_callbacks = self._events_once[event_name]
        try:
            once_event_callbacks.remove(callback)
        except KeyError:
            pass

        event_callbacks = self._events[event_name]
        if not event_callbacks:
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, null_emit_callback)
            elif len(once_event_callbacks) == 1:
                callback = list(once_event_callbacks)[0]

                def emit_callback(*args, **kwargs):
                    setattr(self, "emit_" + event_name, null_emit_callback)
                    once_event_callbacks.clear()
                    return callback(*args, **kwargs)

                setattr(self, "emit_" + event_name, emit_callback)
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        elif len(event_callbacks) == 1:
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, list(event_callbacks)[0])
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        else:
            setattr(self, "emit_" + event_name, self.emit_callback(event_name))

    def remove_listener(self, event_name, callback):
        event_callbacks = self._events[event_name]
        try:
            event_callbacks.remove(callback)
        except KeyError:
            pass

        once_event_callbacks = self._events_once[event_name]
        try:
            once_event_callbacks.remove(callback)
        except KeyError:
            pass

        if not event_callbacks:
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, null_emit_callback)
            elif len(once_event_callbacks) == 1:
                callback = list(once_event_callbacks)[0]

                def emit_callback(*args, **kwargs):
                    setattr(self, "emit_" + event_name, null_emit_callback)
                    once_event_callbacks.clear()
                    return callback(*args, **kwargs)

                setattr(self, "emit_" + event_name, emit_callback)
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        elif len(event_callbacks) == 1:
            if not once_event_callbacks:
                setattr(self, "emit_" + event_name, list(event_callbacks)[0])
            else:
                setattr(self, "emit_" + event_name, self.emit_callback(event_name))
        else:
            setattr(self, "emit_" + event_name, self.emit_callback(event_name))

    def remove_all_listeners(self, event_name=None):
        if event_name is None:
            for event_name in set(list(self._events.keys()) + list(self._events_once.keys())):
                setattr(self, "emit_" + event_name, null_emit_callback)
            self._events = defaultdict(set)
            self._events_once = defaultdict(set)
        else:
            setattr(self, "emit_" + event_name, null_emit_callback)
            self._events[event_name] = set()
            self._events_once[event_name] = set()

    def emit_callback(self, event_name):
        def _(*args, **kwargs):
            for cb in self._events[event_name]:
                try:
                    cb(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, (KeyboardInterrupt, SystemError)):
                        raise e
                    get_logger().exception('error when calling callback:%s',e)

            callbacks = self._events_once[event_name]
            if callbacks:
                self._events_once[event_name] = set()
                while callbacks:
                    cb = callbacks.pop()
                    try:
                        cb(*args, **kwargs)
                    except Exception as e:
                        if isinstance(e, (KeyboardInterrupt, SystemError)):
                            raise e
                        get_logger().exception('error when calling callback:%s',e)
        return _

    def emit(self, event_name, *args, **kwargs):
        for cb in self._events[event_name]:
            try:
                cb(*args, **kwargs)
            except Exception as e:
                if isinstance(e, (KeyboardInterrupt, SystemError)):
                    raise e
                get_logger().exception('error when calling callback:%s', e)

        callbacks = self._events_once[event_name]
        if callbacks:
            self._events_once[event_name] = set()
            while callbacks:
                cb = callbacks.pop()
                try:
                    cb(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, (KeyboardInterrupt, SystemError)):
                        raise e
                    get_logger().exception('error when calling callback:%s', e)

    def __getattr__(self, item):
        if item[:5] == "emit_":
            event_name = item[5:]
            event_callbacks = self._events[event_name]
            if not event_callbacks:
                once_event_callbacks = self._events_once[event_name]
                if not once_event_callbacks:
                    setattr(self, "emit_" + event_name, null_emit_callback)
                    return null_emit_callback
                if len(once_event_callbacks) == 1:
                    callback = list(once_event_callbacks)[0]

                    def emit_callback(*args, **kwargs):
                        setattr(self, "emit_" + event_name, null_emit_callback)
                        once_event_callbacks.clear()
                        return callback(*args, **kwargs)

                    setattr(self, "emit_" + event_name, emit_callback)
                    return emit_callback
                callback = self.emit_callback(event_name)
                setattr(self, "emit_" + event_name, callback)
                return callback

            if len(event_callbacks) == 1:
                once_event_callbacks = self._events_once[event_name]
                if not once_event_callbacks:
                    callback = list(event_callbacks)[0]
                    setattr(self, "emit_" + event_name, callback)
                    return callback
            callback = self.emit_callback(event_name)
            setattr(self, "emit_" + event_name, callback)
            return callback

        elif item[:3] == "on_":
            return lambda *args, **kwargs: self.on(item[3:], *args, **kwargs)
        return object.__getattribute__(self, item)


if is_py3:
    from .coroutines.event import warp_coroutine
    EventEmitter = warp_coroutine(EventEmitter)



# -*- coding: utf-8 -*-

import select
from collections import defaultdict
from ..loop import IOLoop, MODE_NULL, MODE_IN, MODE_OUT, MODE_ERR


class SelectLoop(IOLoop):
    def __init__(self):
        super(SelectLoop, self).__init__()
        self._r_list = set()
        self._w_list = set()
        self._x_list = set()

    def _poll(self, timeout):
        try:
            r, w, x = select.select(self._r_list, self._w_list, self._x_list, min(timeout, 2))
        except Exception as e:
            if isinstance(e, (KeyboardInterrupt, SystemError)):
                raise e
            return []
        results = defaultdict(lambda: MODE_NULL)
        for p in [(r, MODE_IN), (w, MODE_OUT), (x, MODE_ERR)]:
            for fd in p[0]:
                results[fd] |= p[1]
        return results.items()

    def _add_fd(self, fd, mode):
        if mode & MODE_IN:
            self._r_list.add(fd)
        if mode & MODE_OUT:
            self._w_list.add(fd)
        if mode & MODE_ERR:
            self._x_list.add(fd)

    def _remove_fd(self, fd):
        if fd in self._r_list:
            self._r_list.remove(fd)
        if fd in self._w_list:
            self._w_list.remove(fd)
        if fd in self._x_list:
            self._x_list.remove(fd)

    def _modify_fd(self, fd, mode):
        self._remove_fd(fd)
        self._add_fd(fd, mode)



# -*- coding: utf-8 -*-
# 2020/7/10
# create by: snower

import re
from sevent.helpers.tcp2proxy import *

def do_rewrite_host(forward_host, forward_port, rewrite_hosts):
    def parse_rewrite_host(host_args, rewrite_host, rewrite_args):
        for arg in rewrite_args:
            if arg < 0 or arg >= len(host_args):
                continue
            rewrite_host = rewrite_host.replace("{" + str(arg) + "}", host_args[arg])
        rewrite_host = rewrite_host.split(":")
        if len(rewrite_host) == 2:
            return rewrite_host[0], int(rewrite_host[1])
        return rewrite_host[0], forward_port

    for host, (rewrite_host, rewrite_args) in rewrite_hosts:
        if isinstance(host, str):
            if host == "*":
                return parse_rewrite_host(["%s:%s" % (forward_host, forward_port)], rewrite_host, rewrite_args)
            if forward_port in (80, 443):
                if forward_host == host:
                    return parse_rewrite_host([host], rewrite_host, rewrite_args)
            if ("%s:%s" % (forward_host, forward_port)) == host:
                return parse_rewrite_host([host], rewrite_host, rewrite_args)
        else:
            if forward_port in (80, 443):
                matched = host.match(forward_host)
                if matched:
                    return parse_rewrite_host(matched.groups(), rewrite_host, rewrite_args)
            matched = host.match("%s:%s" % (forward_host, forward_port))
            if matched:
                return parse_rewrite_host(matched.groups(), rewrite_host, rewrite_args)
    return forward_host, forward_port

def check_allow_host(forward_host, forward_port, allow_hosts):
    for host in allow_hosts:
        if isinstance(host, str):
            if host == "*":
                return True
            if forward_port in (80, 443):
                if forward_host == host:
                    return True
            if ("%s:%s" % (forward_host, forward_port)) == host:
                return True
        else:
            if forward_port in (80, 443):
                if host.match(forward_host):
                    return True
            if host.match("%s:%s" % (forward_host, forward_port)):
                return True
    return False

def check_noproxy_host(forward_host, forward_port, noproxy_hosts):
    for host in noproxy_hosts:
        if isinstance(host, str):
            if host == "*":
                return True
            if forward_port in (80, 443):
                if forward_host == host:
                    return True
            if ("%s:%s" % (forward_host, forward_port)) == host:
                return True
        else:
            if forward_port in (80, 443):
                if host.match(forward_host):
                    return True
            if host.match("%s:%s" % (forward_host, forward_port)):
                return True
    return False

async def parse_http_forward(conn, rbuffer):
    while True:
        if b'\r\n\r\n' in rbuffer.join():
            break
        rbuffer.write((await conn.recv()).read())

    data = rbuffer.join()
    data = data[:data.index(b'\r\n\r\n')]
    for header in data.split(b'\r\n'):
        try:
            index = header.index(b':')
        except ValueError:
            continue
        if header[:index].decode("utf-8").strip().lower() != "host":
            continue
        host = header[index + 1:].decode("utf-8").strip().split(":")
        if len(host) != 2:
            return header[index + 1:].decode("utf-8").strip(), 80
        return host[0], int(host[1])
    return "", 0

async def parse_tls_forward(conn, rbuffer):
    head_data = (await conn.recv(43)).read(43)
    rbuffer.write(head_data)

    # session_id
    data = (await conn.recv(1)).read(1)
    rbuffer.write(data)
    if ord(data) > 0:
        data = (await conn.recv(ord(data))).read(ord(data))
        rbuffer.write(data)

    # cipher suites
    data = (await conn.recv(2)).read(2)
    rbuffer.write(data)
    csl, = struct.unpack("!H", data)
    if csl > 0:
        data = (await conn.recv(csl)).read(csl)
        rbuffer.write(data)

    # compression methods
    data = (await conn.recv(1)).read(1)
    rbuffer.write(data)
    if ord(data) > 0:
        data = (await conn.recv(ord(data))).read(ord(data))
        rbuffer.write(data)

    # estensions
    data = (await conn.recv(2)).read(2)
    rbuffer.write(data)
    el, = struct.unpack("!H", data)
    if el > 0:
        data = (await conn.recv(el)).read(el)
        rbuffer.write(data)

        i = 0
        while i < len(data):
            et_item, el_item = struct.unpack("!HH", data[i: i + 4])
            if et_item != 0:
                i += el_item + 4
                continue

            sl, = struct.unpack("!H", data[i+4: i+6])
            j, sdata = 0, data[i+6:i+6+sl]
            while j < len(sdata):
                st_item, sl_item = struct.unpack("!BH", sdata[j: j + 3])
                if st_item != 0:
                    j += sl_item + 3
                    continue

                host = sdata[j + 3: j + 3 + sl_item].decode("utf-8").split(":")
                if len(host) != 2:
                    return sdata[j + 3: j + 3 + sl_item].decode("utf-8"), 443
                return host[0], int(host[1])
    return "", 0

async def none_proxy(conns, conn, proxy_host, proxy_port, remote_host, remote_port, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket((remote_host, remote_port))
        await pconn.connectof((remote_host, remote_port))
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("none proxy connected %s:%d -> %s:%d", conn.address[0], conn.address[1], remote_host, remote_port)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("none proxy error %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1], remote_host, remote_port,
                     e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("none proxy closed %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1], remote_host, remote_port,
                 format_data_len(status["send_len"]), format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

async def parse_forward(proxy_type, conns, conn, proxy_host, proxy_port,
                        default_forward_host, default_forward_port, default_forward_proxy_type,
                        allow_hosts, noproxy_hosts, rewrite_hosts, status):
    try:
        rbuffer = sevent.Buffer()
        timer = sevent.current().add_timeout(5, lambda: conn.close())
        try:
            data = (await conn.recv()).read()
            rbuffer.write(data)

            if len(data) < 11 or data[:2] != b'\x16\x03' or data[2] not in (1, 2, 3, 4) or data[5] != 0x01 \
                    or data[9] != 0x03 or data[10] not in (1, 2, 3, 4):
                if data[:data.index(b' ')].decode("utf-8").lower() in ("get", "post", "put", "delete", "head", "option") \
                        or b'HTTP' in data:
                    forward_host, forward_port = await parse_http_forward(conn, rbuffer)
                else:
                    forward_host, forward_port = "", 0
            else:
                rbuffer.write(conn.buffer[0].read())
                conn.buffer[0].write(rbuffer.read())
                forward_host, forward_port = await parse_tls_forward(conn, rbuffer)
        except Exception:
            forward_host, forward_port = "", 0
        finally:
            sevent.current().cancel_timeout(timer)

        if not forward_host or not forward_port:
            forward_host, forward_port, proxy_type = default_forward_host, default_forward_port, default_forward_proxy_type
        elif allow_hosts and allow_hosts[0] != "*" and not check_allow_host(forward_host, forward_port, allow_hosts):
            forward_host, forward_port, proxy_type = default_forward_host, default_forward_port, default_forward_proxy_type
        if rewrite_hosts:
            forward_host, forward_port = do_rewrite_host(forward_host, forward_port, rewrite_hosts)
        if not forward_host or not forward_port:
            logging.info("%s proxy unknown closed %s:%d -> %s:%d", proxy_type, conn.address[0], conn.address[1],
                         proxy_host, proxy_port)
            conn.close()
            conns.pop(id(conn), None)
            return

        rbuffer.write(conn.buffer[0].read())
        conn.buffer[0].write(rbuffer.read())
        if proxy_type == "http":
            if check_noproxy_host(forward_host, forward_port, noproxy_hosts):
                proxy_type = "none"
                await none_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
            else:
                await http_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        elif proxy_type == "socks5":
            if check_noproxy_host(forward_host, forward_port, noproxy_hosts):
                proxy_type = "none"
                await none_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
            else:
                await socks5_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        elif proxy_type == "socks5s":
            if check_noproxy_host(forward_host, forward_port, noproxy_hosts):
                proxy_type = "none"
                await none_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
            else:
                await socks5s_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        else:
            await none_proxy(conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
    except Exception as e:
        logging.info("%s proxy error %s:%d -> %s:%d %s\r%s", proxy_type, conn.address[0], conn.address[1],
                     proxy_host, proxy_port, e, traceback.format_exc())
        conns.pop(id(conn), None)

async def tcp_accept(server, args):
    proxy_info = args.proxy_host.split(":")
    if len(proxy_info) == 1:
        if not proxy_info[0].isdigit():
            proxy_host, proxy_port = proxy_info[0], 8088
        else:
            proxy_host, proxy_port = "127.0.0.1", int(proxy_info[0])
    else:
        proxy_host, proxy_port = proxy_info[0], int(proxy_info[1])

    default_forward_info = args.default_forward_host.split(":")
    if not default_forward_info or not default_forward_info[0]:
        default_forward_host, default_forward_port = "", 0
    elif len(default_forward_info) == 1:
        if not default_forward_info[0].isdigit():
            default_forward_host, default_forward_port = default_forward_info[0], 8088
        else:
            default_forward_host, default_forward_port = "127.0.0.1", int(default_forward_info[0])
    else:
        default_forward_host, default_forward_port = default_forward_info[0], int(default_forward_info[1])
    allow_hosts = []
    for allow_host in (i for i in args.allow_hosts.split(",") if i.strip()):
        if "*" in allow_host and "*" != allow_host:
            allow_hosts.append(re.compile(allow_host.replace(".", "\.").replace("*", ".+?")))
        else:
            allow_hosts.append(allow_host)
    noproxy_hosts = []
    for noproxy_host in (i for i in args.noproxy_hosts.split(",") if i.strip()):
        if "*" in noproxy_host and "*" != noproxy_host:
            noproxy_hosts.append(re.compile(noproxy_host.replace(".", "\.").replace("*", ".+?")))
        else:
            noproxy_hosts.append(noproxy_host)
    rewrite_hosts = []
    for rewrite_host in (i for i in args.rewrite_hosts.split(",") if i.strip()):
        rewrite_host = [i for i in rewrite_host.split("=") if i.strip()]
        if len(rewrite_host) != 2:
            continue
        rewrite_hosts.append((
            re.compile("^" + rewrite_host[0].replace(".", "\.").replace("*", "(.+?)") + "$") if "*" in rewrite_host[0] and "*" != rewrite_host[0] else rewrite_host[0],
            (
                rewrite_host[1],
                tuple((int(i[1:-1]) for i in re.findall("(\{\d+?\})", rewrite_host[1])))
            )
        ))

    logging.info("use %s proxy %s:%d default forward to %s:%d", args.proxy_type, proxy_host, proxy_port,
                 default_forward_host, default_forward_port)
    conns = {}
    sevent.current().call_async(check_timeout, conns, args.timeout)
    while True:
        conn = await server.accept()
        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
        sevent.current().call_async(parse_forward, args.proxy_type, conns, conn, proxy_host, proxy_port,
                                    default_forward_host, default_forward_port, args.default_forward_proxy_type,
                                    allow_hosts, noproxy_hosts, rewrite_hosts, status)
        conns[id(conn)] = (conn, status)

def main(argv):
    parser = argparse.ArgumentParser(description='auto parse servername (support http, TLS) forword to remote host from none, http or socks5 proxy')
    parser.add_argument('-b', dest='bind', default="0.0.0.0", help='local bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='port', default=8088, type=int, help='local bind port (default: 8088)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-T', dest='proxy_type', default="http", choices=("none", "http", "socks5", "socks5s"),
                        help='proxy type (default: http)')
    parser.add_argument('-P', dest='proxy_host', default="127.0.0.1:8088",
                        help='proxy host, accept format [proxy_host:proxy_port] (default: 127.0.0.1:8088)')
    parser.add_argument('-f', dest='default_forward_host', default="",
                        help='default remote forward host , accept format [remote_host:remote_port] (default: )')
    parser.add_argument('-x', dest='default_forward_proxy_type', default="http", choices=("none", "http", "socks5", "socks5s"),
                        help='default remote forward  proxy type (default: http)')
    parser.add_argument('-H', dest='allow_hosts', default="*",
                        help='allow hosts, accept format [host,*host,host*] (default: *)')
    parser.add_argument('-N', dest='noproxy_hosts', default="",
                        help='noproxy hosts, accept format [host,*host,host*] (default: )')
    parser.add_argument('-R', dest='rewrite_hosts', default="",
                        help='rewrite hosts, accept format [host=rhost,*host={1}rhost,host*=rhost{1}] (default: )')
    args = parser.parse_args(args=argv)
    config_signal()
    server = create_server((args.bind, args.port))
    logging.info("listen server at %s:%d", args.bind, args.port)
    sevent.current().call_async(tcp_accept, server, args)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 20/2/14
# create by: snower

import os
import socket
import threading


def set_close_exec(fd):
    import fcntl
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)


def _set_nonblocking(fd):
    import fcntl
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)


class PipeWaker(object):
    def __init__(self):
        r, w = os.pipe()
        _set_nonblocking(r)
        _set_nonblocking(w)
        set_close_exec(r)
        set_close_exec(w)
        self.reader = os.fdopen(r, "rb", 0)
        self.writer = os.fdopen(w, "wb", 0)
        self.lock = threading.Lock()

    def fileno(self):
        return self.reader.fileno()

    def wake(self):
        with self.lock:
            try:
                self.writer.write(b"x")
            except (IOError, ValueError):
                pass

    def consume(self):
        try:
            while True:
                result = self.reader.read()
                if not result:
                    break
        except IOError:
            pass

    def close(self):
        self.reader.close()
        try:
            self.writer.close()
        except Exception:
            pass


class SocketWaker(object):
    def __init__(self):
        self.writer = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.writer.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        count = 0
        while 1:
            count += 1
            a = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            a.bind(("127.0.0.1", 0))
            a.listen(1)
            connect_address = a.getsockname()
            try:
                self.writer.connect(connect_address)
                break
            except socket.error:
                if count >= 10:
                    a.close()
                    self.writer.close()
                    raise socket.error("Cannot bind trigger!")
                a.close()

        self.reader, addr = a.accept()
        self.reader.setblocking(0)
        self.writer.setblocking(0)
        a.close()
        self.lock = threading.Lock()

    def fileno(self):
        return self.reader.fileno()

    def wake(self):
        with self.lock:
            try:
                self.writer.send(b"x")
            except (IOError, socket.error, ValueError):
                pass

    def consume(self):
        try:
            while True:
                result = self.reader.recv(1024)
                if not result or len(result) < 1024:
                    break
        except (IOError, socket.error):
            pass

    def close(self):
        self.reader.close()
        try:
            self.writer.close()
        except Exception:
            pass


def Waker():
    if os.name == 'nt':
        return SocketWaker()
    try:
        return PipeWaker()
    except:
        return SocketWaker()


# -*- coding: utf-8 -*-

version = '0.4.40'
version_info = (0, 4, 40)

from .utils import is_py3, set_logger
from .loop import instance, current
from .event import EventEmitter
from . import tcp
from . import udp
from . import pipe
from .buffer import Buffer
from .dns import DNSResolver
from . import sslsocket as ssl
from . import errors

if is_py3:
    from .coroutines.future import Future
    from .coroutines.chain import Chain
    from . import loop


    def run(callback, *args, **kwargs):
        return loop.instance().run(callback, *args, **kwargs)


    def go(callback, *args, **kwargs):
        if not loop._mul_ioloop:
            return loop._ioloop.go(callback, *args, **kwargs)
        return loop._thread_local._sevent_ioloop.go(callback, *args, **kwargs)


    def sleep(seconds):
        if not loop._mul_ioloop:
            return loop._ioloop.sleep(seconds)
        return loop._thread_local._sevent_ioloop.sleep(seconds)



# -*- coding: utf-8 -*-
# 2021/1/31
# create by: snower

import sys
import os
import time
import struct
import logging
import traceback
import argparse
import threading
import socket
import hashlib
import sevent
from .utils import create_server, create_socket, config_signal, format_data_len
from .simple_proxy import http_protocol_parse, socks5_protocol_parse, socks4_protocol_parse

conns, status = {}, {"remote_conn": [], "local_conn": []}

def warp_write(conn, status, key):
    origin_write = conn.write
    def _(data):
        status[key] += len(data)
        return origin_write(data)
    return _

def gen_sign_key(key):
    t = struct.pack("I", int(time.time()))
    oncestr = os.urandom(16)
    return t + oncestr + hashlib.md5(t + oncestr + sevent.utils.ensure_bytes(key)).digest()

def check_sign_key(key, sign_key):
    return sign_key[20:] == hashlib.md5(sign_key[:20] + sevent.utils.ensure_bytes(key)).digest()

async def parse_forward_address(conn, proxy_type):
    if not proxy_type or proxy_type == "raw":
        return None

    if proxy_type == "socks5" or proxy_type == "socks5s":
        buffer = await conn.recv()
        is_allowed, protocol, host, port, data = await socks5_protocol_parse(conn, buffer)
        if not host or not port:
            raise Exception("parse error")
        if data:
            buffer.write(data + buffer.read())
        return host, port

    if proxy_type == "socks4" or proxy_type == "socks4a":
        buffer = await conn.recv()
        is_allowed, protocol, host, port, data = await socks4_protocol_parse(conn, buffer)
        if not host or not port:
            raise Exception("parse error")
        if data:
            buffer.write(data + buffer.read())
        return host, port

    if proxy_type == "http":
        buffer = await conn.recv()
        is_allowed, host, port, data = await http_protocol_parse(conn, buffer)
        if not host or not port:
            raise Exception("parse error")
        if data:
            buffer.write(data + buffer.read())
        return host, port

    if proxy_type == "redirect":
        address_data = conn.socket.getsockopt(socket.SOL_IP, 80, 16)
        host, port = socket.inet_ntoa(address_data[4:8]), struct.unpack(">H", address_data[2:4])[0]
        return host, port
    raise Exception("parse error")

async def write_forward_address(conn, forward_address):
    await conn.send(b"".join([struct.pack("!H", len(forward_address[0])),
                               sevent.utils.ensure_bytes(forward_address[0]), struct.pack("!H", forward_address[1])]))

async def read_forward_address(conn):
    host_len, = struct.unpack("!H", (await conn.recv(2)).read(2))
    host = sevent.utils.ensure_unicode((await conn.recv(host_len)).read(host_len))
    port, = struct.unpack("!H", (await conn.recv(2)).read(2))
    return host, port

async def tcp_forward(conn, forward_address, conns, status, conn_status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket(forward_address)
        await pconn.connectof(forward_address)
        if conn in conn_status["remote_conn"]:
            conn_status["remote_conn"].remove(conn)
        conns[id(conn)] = (conn, pconn, status)
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("tcp forward connected %s:%d -> %s:%d", conn.address[0], conn.address[1], forward_address[0], forward_address[1])
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("tcp forward error %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1],
                     forward_address[0], forward_address[1], e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("tcp forward closed %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1],
                 forward_address[0], forward_address[1], format_data_len(status["send_len"]), format_data_len(status["recv_len"]),
                 (time.time() - start_time) * 1000)

async def reverse_port_forward(remote_conn, local_conn, status, forward_address):
    start_time = time.time()

    try:
        if forward_address:
            await remote_conn.send(b'\x01')
            await write_forward_address(remote_conn, forward_address)
        else:
            await remote_conn.send(b'\x00')
        local_conn.write = warp_write(local_conn, status, "recv_len")
        remote_conn.write = warp_write(remote_conn, status, "send_len")
        logging.info("tcp forward connected %s:%d -> %s:%d", local_conn.address[0], local_conn.address[1],
                     remote_conn.address[0], remote_conn.address[1])
        await local_conn.linkof(remote_conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("tcp forward error %s:%d -> %s:%d %s %.2fms\r%s", local_conn.address[0], local_conn.address[1],
                     remote_conn.address[0], remote_conn.address[1], e, (time.time() - start_time) * 1000,
                     traceback.format_exc())
        return
    finally:
        remote_conn.close()
        local_conn.close()
        conns.pop(id(remote_conn), None)

    logging.info("tcp forward closed %s:%d -> %s:%d %s %s %.2fms", local_conn.address[0], local_conn.address[1],
                 remote_conn.address[0], remote_conn.address[1], format_data_len(status["send_len"]),
                 format_data_len(status["recv_len"]),
                 (time.time() - start_time) * 1000)

async def server_handle_remote_connect(conn, forward_address, key, proxy_type, conns, status):
    setattr(conn, "_connected_time", time.time())
    def on_close(conn):
        if conn not in status["remote_conn"]:
            return
        status["remote_conn"].remove(conn)
        logging.info("remote conn waited close %s:%d", conn.address[0], conn.address[1])

    status["remote_conn"].append(conn)
    conn.on_close(on_close)
    try:
        command_type, sign_key_len = struct.unpack("!BB", (await conn.recv(2)).read(2))
        sign_key = (await conn.recv(sign_key_len)).read(sign_key_len) if sign_key_len > 0 else b''
    except sevent.errors.SocketClosed:
        return
    if not check_sign_key(key, sign_key):
        await conn.closeof()
        logging.info("remote conn auth fail %s:%d %s", conn.address[0], conn.address[1], sign_key)
        return

    if command_type == 1:
        if status["local_conn"]:
            forward_status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0,
                              "check_send_len": 0}
            local_conn = status["local_conn"].pop(0)
            if conn in status["remote_conn"]:
                status["remote_conn"].remove(conn)
            conns[id(conn)] = (conn, local_conn, forward_status)
            sevent.go(reverse_port_forward, conn, local_conn, forward_status, local_conn._connected_forward_address)
            return
        setattr(conn, "_heartbeat_time", time.time())
        logging.info("remote conn waiting %s:%d", conn.address[0], conn.address[1])
        return

    if command_type == 0x02 or command_type == 0x03:
        try:
            if command_type == 0x03:
                forward_address = await read_forward_address(conn)
            await conn.send(b'\x00')
            forward_status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0,
                              "check_send_len": 0}
            sevent.current().call_async(tcp_forward, conn, forward_address, conns, forward_status, status)
        except sevent.errors.SocketClosed:
            pass
        except Exception as e:
            logging.info("tcp forward error %s:%d -> %s:%d %s\r%s", conn.address[0], conn.address[1],
                         forward_address[0], forward_address[1], e, traceback.format_exc())
        return

    await conn.closeof()
    logging.info("remote conn unsupport connect type %s:%d %s", conn.address[0], conn.address[1], key)

async def server_handle_local_connect(conn, forward_address, key, proxy_type, conns, status):
    try:
        forward_address = (await parse_forward_address(conn, proxy_type)) if proxy_type else forward_address
    except Exception as e:
        logging.info("parse proxy forward address error %s", e)
        return

    setattr(conn, "_connected_forward_address", forward_address)
    setattr(conn, "_connected_time", time.time())
    if status["remote_conn"]:
        forward_status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0,
                          "check_send_len": 0}
        remote_conn = status["remote_conn"].pop(0)
        conns[id(remote_conn)] = (remote_conn, conn, forward_status)
        sevent.go(reverse_port_forward, remote_conn, conn, forward_status, forward_address)
        return

    def on_close(conn):
        if conn not in status["local_conn"]:
            return
        status["local_conn"].remove(conn)
        logging.info("local conn waited close %s:%d", conn.address[0], conn.address[1])
    status["local_conn"].append(conn)
    conn.on_close(on_close)
    logging.info("local conn waiting %s:%d", conn.address[0], conn.address[1])

async def server_handle_remote_heartbeat(conn, conn_status):
    if conn not in conn_status["remote_conn"]:
        return
    conn_status["remote_conn"].remove(conn)
    try:
        await conn.send(b'\x04')
        command_type = (await conn.recv(1)).read(1)
        if command_type != b'\x04':
            sevent.current().add_async(conn.close)
        else:
            setattr(conn, "_heartbeat_time", time.time())
            if status["local_conn"]:
                forward_status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0,
                                  "check_send_len": 0}
                local_conn = status["local_conn"].pop(0)
                if conn in status["remote_conn"]:
                    status["remote_conn"].remove(conn)
                conns[id(conn)] = (conn, local_conn, forward_status)
                sevent.go(reverse_port_forward, conn, local_conn, forward_status, local_conn._connected_forward_address)
                return
            conn_status["remote_conn"].append(conn)
    except:
        conn.close()

async def server_run_server(server, forward_address, key, proxy_type, conns, status, handle):
    while True:
        try:
            conn = await server.accept()
            sevent.current().call_async(handle, conn, forward_address, key, proxy_type, conns, status)
        except sevent.errors.SocketClosed as e:
            sevent.current().call_async(sevent.current().stop)
            raise e

async def client_run_connect(remote_address, forward_address, key, conns, status):
    while True:
        start_time = time.time()
        try:
            conn = create_socket(remote_address)
            await conn.connectof(remote_address)
            sign_key = gen_sign_key(key)
            await conn.send(struct.pack("!BB", 0x01, len(sign_key)) + sign_key)
            while True:
                command_type = (await conn.recv(1)).read(1)
                if command_type == b'\x01':
                    current_forward_address = await read_forward_address(conn)
                    break
                elif command_type == b'\x04':
                    setattr(conn, "_heartbeat_time", time.time())
                    await conn.send(b'\x04')
                else:
                    current_forward_address = forward_address
                    break
            forward_status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0,
                              "check_send_len": 0}
            sevent.current().call_async(tcp_forward, conn, current_forward_address, conns, forward_status, status)
        except sevent.errors.SocketClosed as e:
            logging.info("connect error %s:%d %s", remote_address[0], remote_address[1], e)
            if time.time() - start_time < 5:
                await sevent.sleep(5)
        except (sevent.errors.ResolveError, sevent.errors.ConnectTimeout, sevent.errors.ConnectError, ConnectionRefusedError) as e:
            logging.info("connect error %s:%d %s", remote_address[0], remote_address[1], e)
            await sevent.sleep(5)
        except Exception as e:
            sevent.current().call_async(sevent.current().stop)
            raise e

async def client_handle_local_connect(conn, remote_address, key, proxy_type, conns, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        forward_address = (await parse_forward_address(conn, proxy_type)) if proxy_type else None

        pconn = create_socket(remote_address)
        await pconn.connectof(remote_address)
        sign_key = gen_sign_key(key)
        if forward_address:
            await pconn.send(struct.pack("!BB", 0x03, len(sign_key)) + sign_key)
            await write_forward_address(pconn, forward_address)
        else:
            await pconn.send(struct.pack("!BB", 0x02, len(sign_key)) + sign_key)
        (await pconn.recv(1)).read(1)
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("tcp forward connected %s:%d -> %s:%d", conn.address[0], conn.address[1], remote_address[0], remote_address[1])
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("tcp forward error %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1],
                     remote_address[0], remote_address[1], e, (time.time() - start_time) * 1000,
                     traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("tcp forward closed %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1],
                 remote_address[0], remote_address[1], format_data_len(status["send_len"]),
                 format_data_len(status["recv_len"]),
                 (time.time() - start_time) * 1000)

async def client_run_server(server, remote_address, key, proxy_type, conns):
    while True:
        try:
            conn = await server.accept()
            status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
            sevent.current().call_async(client_handle_local_connect, conn, remote_address, key, proxy_type, conns, status)
            conns[id(conn)] = (conn, conn, status)
        except sevent.errors.SocketClosed as e:
            sevent.current().call_async(sevent.current().stop)
            raise e

async def check_timeout(conns, conn_status, timeout):
    def run_check():
        while True:
            try:
                now = time.time()
                for conn in tuple(conn_status["remote_conn"]):
                    if not hasattr(conn, "_heartbeat_time"):
                        if now - conn._connected_time >= 30:
                            sevent.current().add_async_safe(conn.close)
                    else:
                        if now - conn._heartbeat_time >= 90:
                            if now - conn._heartbeat_time >= 180:
                                sevent.current().add_async_safe(conn.close)
                            elif not hasattr(conn, "_ping_time") or now - conn._ping_time >= 90:
                                setattr(conn, "_ping_time", time.time())
                                sevent.current().add_async_safe(lambda c, s: sevent.go(server_handle_remote_heartbeat, c, s),
                                                                conn, conn_status)

                for conn in tuple(conn_status["local_conn"]):
                    if now - conn._connected_time >= 60:
                        sevent.current().add_async_safe(conn.close)

                for conn_id, (conn, pconn, status) in list(conns.items()):
                    if status['check_recv_len'] != status['recv_len'] or status['check_send_len'] != status['send_len']:
                        status["check_recv_len"] = status["recv_len"]
                        status["check_send_len"] = status["send_len"]
                        status['last_time'] = now
                        continue

                    if now - status['last_time'] >= timeout:
                        sevent.current().add_async_safe(conn.close)
                        sevent.current().add_async_safe(pconn.close)
                        conns.pop(conn_id, None)
            finally:
                time.sleep(min(float(timeout) / 2.0, 30))

    if timeout > 0:
        check_thread = threading.Thread(target=run_check)
        check_thread.daemon = True
        check_thread.start()
    await sevent.Future()

def main(argv):
    parser = argparse.ArgumentParser(description='tcp reverse port forward')
    parser.add_argument('-c', dest='is_client_mode', nargs='?', const=True, default=False, type=bool,
                        help='is client mode (defualt: False)')
    parser.add_argument('-k', dest='key', default='', type=str,
                        help='auth key (defualt: "")')
    parser.add_argument('-b', dest='bind_host', default="0.0.0.0",
                        help='server and client mode local bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='bind_port', default=0, type=int,
                        help='server and client mode local bind port (default: 8089)')
    parser.add_argument('-r', dest='listen_host', default="0.0.0.0",
                        help='server mode reverse server listen host (default: 0.0.0.0)')
    parser.add_argument('-l', dest='listen_port', default=8088, type=int,
                        help='server mode reverse server listen port (default: 8088)')
    parser.add_argument('-H', dest='connect_host', default="127.0.0.1",
                        help='client mode reverse client connect server host (default: 127.0.0.1)')
    parser.add_argument('-P', dest='connect_port', default=8088, type=int,
                        help='client mode reverse client connect server port (default: 8088)')
    parser.add_argument('-f', dest='forward_host', default="",
                        help='server and client mode forward host , accept format [remote_host:remote_port] (default: )')
    parser.add_argument('-T', dest='proxy_type', default="",
                        choices=("raw", "http", "socks4", "socks4a", "socks5", "socks5s", "redirect"),
                        help='server and client mode local listen proxy type (default: raw)')
    parser.add_argument('-t', dest='timeout', default=7200,
                        type=int, help='no read/write timeout (default: 7200)')
    args = parser.parse_args(args=argv)
    config_signal()

    if not args.forward_host:
        forward_address = None
    else:
        forward_info = args.forward_host.split(":")
        if len(forward_info) == 1:
            if not forward_info[0].isdigit():
                forward_address = (forward_info[0], 8088)
            else:
                forward_address = ("127.0.0.1", int(forward_info[0]))
        else:
            forward_address = (forward_info[0], int(forward_info[1]))

    if not args.is_client_mode:
        remote_server = create_server((args.listen_host, args.listen_port))
        local_server = create_server((args.bind_host, args.bind_port or 8089))
        logging.info("listen %s %d -> %s:%d", args.bind_host, args.bind_port or 8089, args.listen_host, args.listen_port)

        sevent.instance().call_async(server_run_server, remote_server,
                                     forward_address if forward_address else ("127.0.0.1", 80),
                                     sevent.utils.ensure_bytes(args.key), args.proxy_type,
                                     conns, status, server_handle_remote_connect)
        sevent.instance().call_async(server_run_server, local_server, forward_address,
                                     sevent.utils.ensure_bytes(args.key), args.proxy_type,
                                     conns, status, server_handle_local_connect)
    else:
        if args.bind_port:
            local_server = create_server((args.bind_host, args.bind_port))
            logging.info("listen %s %d", args.bind_host, args.bind_port)
            sevent.instance().call_async(client_run_server, local_server, (args.connect_host, args.connect_port),
                                         sevent.utils.ensure_bytes(args.key), args.proxy_type, conns)

        logging.info("connect %s:%d -> %s", args.connect_host, args.connect_port, forward_address)
        sevent.instance().call_async(client_run_connect, (args.connect_host, args.connect_port),
                                     forward_address if forward_address else ("127.0.0.1", 80),
                                     sevent.utils.ensure_bytes(args.key), conns, status)
    sevent.current().call_async(check_timeout, conns, status, args.timeout)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2020/5/8
# create by: snower

import types
import greenlet
from ..utils import get_logger


def warp_coroutine(BaseIOLoop):
    class IOLoop(BaseIOLoop):
        def call_async(self, callback, *args, **kwargs):
            if isinstance(callback, types.CoroutineType):
                def run_coroutine_fuc(*args, **kwargs):
                    def run_coroutine():
                        try:
                            callback.send(None)
                            while True:
                                callback.send(None)
                        except StopIteration:
                            return
                        except Exception as e:
                            if isinstance(e, (KeyboardInterrupt, SystemError)):
                                raise e
                            get_logger().exception("loop callback error:%s", e)

                    child_gr = greenlet.greenlet(run_coroutine)
                    return child_gr.switch()
                return self._handlers.append((run_coroutine_fuc, args, kwargs))

            if callback.__code__.co_flags & 0x80 == 0:
                return self._handlers.append((callback, args, kwargs))

            def run_async_fuc(*args, **kwargs):
                def run_async():
                    try:
                        g = callback(*args, **kwargs)
                        g.send(None)
                        while True:
                            g.send(None)
                    except StopIteration:
                        return
                    except Exception as e:
                        if isinstance(e, (KeyboardInterrupt, SystemError)):
                            raise e
                        get_logger().exception("loop callback error:%s", e)
                child_gr = greenlet.greenlet(run_async)
                return child_gr.switch()
            return self._handlers.append((run_async_fuc, args, kwargs))

        go = call_async

        async def sleep(self, seconds):
            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"
            self.add_timeout(seconds, child_gr.switch)
            return main.switch()

        def run(self, callback, *args, **kwargs):
            if isinstance(callback, types.CoroutineType):
                async def do_coroutine_run():
                    try:
                        await callback
                    finally:
                        self.stop()
                self.call_async(do_coroutine_run)
                return self.start()

            if callback.__code__.co_flags & 0x80 == 0:
                def do_run():
                    try:
                        callback(*args, **kwargs)
                    finally:
                        self.stop()
                self.add_async(do_run)
                return self.start()

            async def do_async_run():
                try:
                    await callback(*args, **kwargs)
                finally:
                    self.stop()
            self.call_async(do_async_run)
            return self.start()

    return IOLoop



# -*- coding: utf-8 -*-
# 2020/7/10
# create by: snower

import re
import sys
import time
import argparse
import logging
import traceback
import threading
import sevent
from .utils import create_server, create_socket, config_signal
from .simple_proxy import format_data_len, warp_write, http_protocol_parse, socks5_protocol_parse, socks4_protocol_parse
from .tcp2proxy import http_build_protocol, socks5_build_protocol, socks5_read_protocol

def check_host(forward_host, allow_hosts):
    for host in allow_hosts:
        if isinstance(host, str):
            if forward_host == host or forward_host.endswith(host):
                return True
        else:
            if host.match(forward_host):
                return True
    return False

async def socks5_proxy(proxy_host, proxy_port, remote_host, remote_port):
    pconn = None
    try:
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))
        await pconn.send(b"\x05\x01\x00")
        buffer = await pconn.recv()
        if buffer.read() != b'\x05\00':
            logging.info("protocol hello error")
            return

        protocol_data = socks5_build_protocol(remote_host, remote_port)
        await pconn.send(protocol_data)
        buffer = await pconn.recv()
        if buffer.read(3) != b'\x05\x00\x00':
            logging.info("protocol error")
            return
        if not socks5_read_protocol(buffer):
            logging.info("protocol error")
            return
    except sevent.errors.SocketClosed:
        pconn = None
    except Exception as e:
        if pconn: pconn.close()
        raise e
    return pconn

async def socks5s_proxy(proxy_host, proxy_port, remote_host, remote_port):
    pconn = None
    try:
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))
        await pconn.send(b"\x05\x01\x8e" + socks5_build_protocol(remote_host, remote_port))
    except sevent.errors.SocketClosed:
        pconn = None
    except Exception as e:
        if pconn: pconn.close()
        raise e
    return pconn

async def http_proxy(proxy_host, proxy_port, remote_host, remote_port):
    pconn = None
    try:
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))

        protocol_data = http_build_protocol(remote_host, remote_port)
        await pconn.send(protocol_data)
        buffer = await pconn.recv()
        if buffer.read(12).lower() != b"http/1.1 200":
            logging.info("protocol error")
            return
        buffer.read()
    except sevent.errors.SocketClosed:
        pconn = None
    except Exception as e:
        if pconn: pconn.close()
        raise e
    return pconn

async def none_proxy(proxy_host, proxy_port, remote_host, remote_port):
    pconn = None
    try:
        pconn = create_socket((remote_host, remote_port))
        await pconn.connectof((remote_host, remote_port))
    except sevent.errors.SocketClosed:
        pconn = None
    except Exception as e:
        if pconn: pconn.close()
        raise e
    return pconn

async def tcp_proxy(conns, conn, proxy_type, proxy_host, proxy_port, noproxy_hosts, status):
    start_time = time.time()
    host, port, protocol = '', 0, ''
    conn.write, pconn = warp_write(conn, status, "recv_len"), None
    try:
        conn.enable_nodelay()
        buffer = await conn.recv()
        if buffer[0] == 5:
            is_allowed, protocol, host, port, data = await socks5_protocol_parse(conn, buffer)
        elif buffer[0] == 4:
            is_allowed, protocol, host, port, data = await socks4_protocol_parse(conn, buffer)
        else:
            protocol = 'http'
            is_allowed, host, port, data = await http_protocol_parse(conn, buffer)
        if not host or not port:
            logging.info("empty address")
            return

        logging.info("connected %s %s:%d -> %s %s:%d -> %s:%d", protocol, conn.address[0], conn.address[1],
                     proxy_type, proxy_host, proxy_port, host, port)
        if check_host(host, noproxy_hosts):
            proxy_type = "none"
            pconn = await none_proxy(proxy_host, proxy_port, host, port)
        else:
            if proxy_type == "http":
                pconn = await http_proxy(proxy_host, proxy_port, host, port)
            elif proxy_type == "socks5s":
                pconn = await socks5s_proxy(proxy_host, proxy_port, host, port)
            else:
                pconn = await socks5_proxy(proxy_host, proxy_port, host, port)
        pconn.write = warp_write(pconn, status, "send_len")
        await pconn.connectof((host, port))
        if data:
            await pconn.send(data)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("error %s %s:%d -> %s %s:%d -> %s:%d %s %.2fms\r%s", protocol, conn.address[0], conn.address[1],
                     proxy_type, proxy_host, proxy_port, host, port, e,
                     (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("closed %s %s:%d -> %s %s:%d -> %s:%d %s %s %.2fms", protocol, conn.address[0], conn.address[1],
                 proxy_type, proxy_host, proxy_port, host, port, format_data_len(status["send_len"]),
                 format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

async def check_timeout(conns, timeout):
    def run_check():
        while True:
            try:
                now = time.time()
                for conn_id, (conn, status) in list(conns.items()):
                    if status['check_recv_len'] != status['recv_len'] or status['check_send_len'] != status['send_len']:
                        status["check_recv_len"] = status["recv_len"]
                        status["check_send_len"] = status["send_len"]
                        status['last_time'] = now
                        continue

                    if now - status['last_time'] >= timeout:
                        sevent.current().add_async_safe(conn.close)
                        conns.pop(conn_id, None)
            finally:
                time.sleep(min(float(timeout) / 2.0, 30))

    if timeout > 0:
        check_thread = threading.Thread(target=run_check)
        check_thread.daemon = True
        check_thread.start()
    await sevent.Future()

async def tcp_accept(server, args):
    proxy_info = args.proxy_host.split(":")
    if len(proxy_info) == 1:
        if not proxy_info[0].isdigit():
            proxy_host, proxy_port = proxy_info[0], 8088
        else:
            proxy_host, proxy_port = "127.0.0.1", int(proxy_info[0])
    else:
        proxy_host, proxy_port = proxy_info[0], int(proxy_info[1])
    noproxy_hosts = []
    for noproxy_host in args.noproxy_hosts.split(","):
        if "*" in noproxy_host and "*" != noproxy_host:
            noproxy_hosts.append(re.compile(noproxy_host.replace(".", "\.").replace("*", ".+?")))
        else:
            noproxy_hosts.append(noproxy_host)

    logging.info("use %s proxy %s:%d", args.proxy_type, proxy_host, proxy_port)
    conns = {}
    sevent.current().call_async(check_timeout, conns, args.timeout)
    while True:
        conn = await server.accept()
        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
        sevent.current().call_async(tcp_proxy, conns, conn, args.proxy_type, proxy_host, proxy_port, noproxy_hosts, status)
        conns[id(conn)] = (conn, status)

def main(argv):
    parser = argparse.ArgumentParser(description='simple http and socks5 proxy forward to http or socks5 uplink proxy')
    parser.add_argument('-b', dest='bind', default="0.0.0.0", help='local bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='port', default=8088, type=int, help='local bind port (default: 8088)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-T', dest='proxy_type', default="http", choices=("http", "socks5", "socks5s"),
                        help='proxy type (default: http)')
    parser.add_argument('-P', dest='proxy_host', default="127.0.0.1:8088",
                        help='proxy host, accept format [proxy_host:proxy_port]  (default: 127.0.0.1:8088)')
    parser.add_argument('-N', dest='noproxy_hosts', default="*",
                        help='noproxy hosts, accept format [host,*host,host*] (default: *)')
    args = parser.parse_args(args=argv)
    config_signal()
    server = create_server((args.bind, args.port))
    logging.info("listen server at %s:%d", args.bind, args.port)
    sevent.run(tcp_accept, server, args)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2020/1/16
# create by: snower


class SeventException(Exception):
    pass


class SocketClosed(SeventException):
    pass

class ChainClosed(SeventException):
    pass

class ResolveError(SeventException):
    pass


class ConnectTimeout(SeventException):
    pass


class AddressError(SeventException):
    pass


class ConnectError(SeventException):
    def __init__(self, address, socket_error, *args, **kwargs):
        super(ConnectError, self).__init__(*args, **kwargs)

        self.address = address
        self.socket_error = socket_error


class SSLConnectError(ConnectError):
    pass


class SSLSocketError(SeventException):
    pass


#!/usr/bin/env python

import os
import sys
import platform
from setuptools import setup, Extension

if platform.python_implementation() == "CPython":
    if platform.system() != 'Windows':
        ext_modules = [Extension('sevent.cbuffer', sources=['sevent/cbuffer.c'])]
    else:
        if sys.version_info[0] >= 3:
            ext_modules = [Extension('sevent.cbuffer', sources=['sevent/cbuffer.c'], libraries=["ws2_32"])]
        else:
            ext_modules = []
else:
    ext_modules = []

if os.path.exists("README.md"):
    if sys.version_info[0] >= 3:
        with open("README.md", encoding="utf-8") as fp:
            long_description = fp.read()
    else:
        with open("README.md") as fp:
            long_description = fp.read()
else:
    long_description = ''

setup(
    name='sevent',
    version='0.4.40',
    packages=['sevent', 'sevent.impl', 'sevent.coroutines', 'sevent.sslsocket', 'sevent.helpers'],
    ext_modules=ext_modules,
    package_data={
        '': ['README.md'],
    },
    install_requires=[
        'dnslib>=0.9.7',
        'greenlet>=0.4.2',
    ],
    author='snower',
    author_email='sujian199@gmail.com',
    url='https://github.com/snower/sevent',
    license='MIT',
    description='lightweight event loop',
    long_description=long_description,
    long_description_content_type="text/markdown",
)



# -*- coding: utf-8 -*-
# 2020/7/10
# create by: snower

import sys
import time
import logging
import struct
import socket
import traceback
import argparse
import threading
import sevent
from .utils import create_server, create_socket, config_signal, format_data_len

def warp_write(conn, status, key):
    origin_write = conn.write
    def _(data):
        status[key] += len(data)
        return origin_write(data)
    return _

def socks5_build_protocol(remote_host, remote_port):
    try:
        protocol_data = b"".join(
            [b"\x05\x01\x00\x01", socket.inet_aton(remote_host), struct.pack(">H", remote_port)])
    except:
        try:
            protocol_data = b"".join([b"\x05\x01\x00\x04", socket.inet_pton(socket.AF_INET6, remote_host),
                                      struct.pack(">H", remote_port)])
        except:
            protocol_data = b"".join([b"\x05\x01\x00\x03", struct.pack(">B", len(remote_host)),
                                      bytes(remote_host, "utf-8") if isinstance(remote_host, str) else remote_host,
                                      struct.pack(">H", remote_port)])
    return protocol_data

def socks5_read_protocol(buffer):
    cmd = buffer.read(1)
    if cmd == b'\x01':
        return buffer.read(6)
    if cmd == b'\x04':
        return buffer.read(18)
    if cmd == b'\x03':
        addr_len = ord(buffer.read(1))
        return buffer.read(addr_len + 2)
    return False

async def socks5_proxy(conns, conn, proxy_host, proxy_port, remote_host, remote_port, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))
        await pconn.send(b"\x05\x01\x00")
        buffer = await pconn.recv()
        if buffer.read() != b'\x05\00':
            logging.info("protocol hello error")
            return

        protocol_data = socks5_build_protocol(remote_host, remote_port)
        await pconn.send(protocol_data)
        buffer = await pconn.recv()
        if buffer.read(3) != b'\x05\x00\x00':
            logging.info("protocol error")
            return
        if not socks5_read_protocol(buffer):
            logging.info("protocol error")
            return
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("socks5 proxy connected %s:%d -> %s:%d -> %s:%d", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("socks5 proxy error %s:%d -> %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port, e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("socks5 proxy closed %s:%d -> %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1], proxy_host, proxy_port,
                 remote_host, remote_port, format_data_len(status["send_len"]), format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

async def socks5s_proxy(conns, conn, proxy_host, proxy_port, remote_host, remote_port, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))
        await pconn.send(b"\x05\x01\x8e" + socks5_build_protocol(remote_host, remote_port))
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("socks5s proxy connected %s:%d -> %s:%d -> %s:%d", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("socks5s proxy error %s:%d -> %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port, e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("socks5s proxy closed %s:%d -> %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1], proxy_host, proxy_port,
                 remote_host, remote_port, format_data_len(status["send_len"]), format_data_len(status["recv_len"]), (time.time() - start_time) * 1000)

def http_build_protocol(remote_host, remote_port):
    remote = (bytes(remote_host, "utf-8") if isinstance(remote_host, str) else remote_host) + b':' + bytes(str(remote_port), "utf-8")
    return b"CONNECT " + remote + b" HTTP/1.1\r\nHost: " + remote + b"\r\nUser-Agent: sevent\r\nProxy-Connection: Keep-Alive\r\n\r\n"

async def http_proxy(conns, conn, proxy_host, proxy_port, remote_host, remote_port, status):
    start_time = time.time()
    conn.write, pconn = warp_write(conn, status, "recv_len"), None

    try:
        conn.enable_nodelay()
        pconn = create_socket((proxy_host, proxy_port))
        await pconn.connectof((proxy_host, proxy_port))

        protocol_data = http_build_protocol(remote_host, remote_port)
        await pconn.send(protocol_data)
        buffer = await pconn.recv()
        if buffer.read(12).lower() != b"http/1.1 200":
            logging.info("protocol error")
            return
        buffer.read()
        pconn.write = warp_write(pconn, status, "send_len")
        logging.info("tcp2proxy connected %s:%d -> %s:%d -> %s:%d", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port)
        await pconn.linkof(conn)
    except sevent.errors.SocketClosed:
        pass
    except Exception as e:
        logging.info("tcp2proxy error %s:%d -> %s:%d -> %s:%d %s %.2fms\r%s", conn.address[0], conn.address[1], proxy_host, proxy_port,
                     remote_host, remote_port, e, (time.time() - start_time) * 1000, traceback.format_exc())
        return
    finally:
        conn.close()
        if pconn: pconn.close()
        conns.pop(id(conn), None)

    logging.info("tcp2proxy closed %s:%d -> %s:%d -> %s:%d %s %s %.2fms", conn.address[0], conn.address[1], proxy_host, proxy_port,
                 remote_host, remote_port, format_data_len(status["send_len"]), format_data_len(status["recv_len"]),
                 (time.time() - start_time) * 1000)

async def check_timeout(conns, timeout):
    def run_check():
        while True:
            try:
                now = time.time()
                for conn_id, (conn, status) in list(conns.items()):
                    if status['check_recv_len'] != status['recv_len'] or status['check_send_len'] != status['send_len']:
                        status["check_recv_len"] = status["recv_len"]
                        status["check_send_len"] = status["send_len"]
                        status['last_time'] = now
                        continue

                    if now - status['last_time'] >= timeout:
                        sevent.current().add_async_safe(conn.close)
                        conns.pop(conn_id, None)
            finally:
                time.sleep(min(float(timeout) / 2.0, 30))

    if timeout > 0:
        check_thread = threading.Thread(target=run_check)
        check_thread.daemon = True
        check_thread.start()
    await sevent.Future()

async def tcp_accept(server, args):
    proxy_info = args.proxy_host.split(":")
    if len(proxy_info) == 1:
        if not proxy_info[0].isdigit():
            proxy_host, proxy_port = proxy_info[0], 8088
        else:
            proxy_host, proxy_port = "127.0.0.1", int(proxy_info[0])
    else:
        proxy_host, proxy_port = proxy_info[0], int(proxy_info[1])

    forward_info = args.forward_host.split(":")
    if len(forward_info) == 1:
        if not forward_info[0].isdigit():
            forward_host, forward_port = forward_info[0], 8088
        else:
            forward_host, forward_port = "127.0.0.1", int(forward_info[0])
    else:
        forward_host, forward_port = forward_info[0], int(forward_info[1])

    logging.info("use %s proxy %s:%d forward to %s:%d", args.proxy_type, proxy_host, proxy_port, forward_host, forward_port)
    conns = {}
    sevent.current().call_async(check_timeout, conns, args.timeout)
    while True:
        conn = await server.accept()
        status = {"recv_len": 0, "send_len": 0, "last_time": time.time(), "check_recv_len": 0, "check_send_len": 0}
        if args.proxy_type == "http":
            sevent.current().call_async(http_proxy, conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        elif args.proxy_type == "socks5s":
            sevent.current().call_async(socks5s_proxy, conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        else:
            sevent.current().call_async(socks5_proxy, conns, conn, proxy_host, proxy_port, forward_host, forward_port, status)
        conns[id(conn)] = (conn, status)

def main(argv):
    parser = argparse.ArgumentParser(description='forword tcp port to remote host from http or socks5 proxy')
    parser.add_argument('-b', dest='bind', default="0.0.0.0", help='local bind host (default: 0.0.0.0)')
    parser.add_argument('-p', dest='port', default=8088, type=int, help='local bind port (default: 8088)')
    parser.add_argument('-t', dest='timeout', default=7200, type=int, help='no read/write timeout (default: 7200)')
    parser.add_argument('-T', dest='proxy_type', default="http", choices=("http", "socks5", "socks5s"), help='proxy type (default: http)')
    parser.add_argument('-P', dest='proxy_host', default="127.0.0.1:8088", help='proxy host, accept format [proxy_host:proxy_port] (default: 127.0.0.1:8088)')
    parser.add_argument('-f', dest='forward_host', default="127.0.0.1:80", help='remote forward host , accept format [remote_host:remote_port] (default: 127.0.0.1:80)')
    args = parser.parse_args(args=argv)
    config_signal()
    server = create_server((args.bind, args.port))
    logging.info("listen server at %s:%d", args.bind, args.port)
    sevent.current().call_async(tcp_accept, server, args)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)1.1s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')
    try:
        main(sys.argv[1:])
        sevent.instance().start()
    except KeyboardInterrupt:
        exit(0)


# -*- coding: utf-8 -*-
# 2020/5/8
# create by: snower

import greenlet


def warp_coroutine(BaseDNSResolver):
    class DNSResolver(BaseDNSResolver):
        async def gethostbyname(self, hostname, timeout=None):
            child_gr = greenlet.getcurrent()
            main = child_gr.parent
            assert main is not None, "must be running in async func"
            self.resolve(hostname, lambda hostname, ip: child_gr.switch(ip), timeout)
            return main.switch()

    return DNSResolver



# -*- coding: utf-8 -*-
# 18/5/25
# create by: snower

import sys
import logging

if sys.version_info[0] >= 3:
    is_py3 = True
    unicode_type = str
    byte_type = bytes

    def is_int(v):
        return v.__class__ is int

    iter_range = range
else:
    is_py3 = False
    unicode_type = unicode
    byte_type = str


    def is_int(v):
        return v.__class__ is int or v.__class__ is long

    iter_range = xrange


def ensure_bytes(s):
    if isinstance(s, unicode_type):
        return s.encode("utf-8")
    return s


def ensure_unicode(s):
    if isinstance(s, byte_type):
        return s.decode("utf-8")
    return s

_logger = logging

def set_logger(logger):
    global _logger
    _logger = logger

def get_logger():
    return _logger